-- Test Bench for IIR filter
--
library ieee;
use std.textio.all;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use work.iir_types.all;

entity iir_test_bench is
end entity iir_test_bench;


architecture test of iir_test_bench is
   -- Path
   constant path : string := "C:\User\Daniel\Protel\FPGA_Biquad\";
   -- File with IIR filter coefficients
   constant filter_coeff : string := "filters\iir_coeff_2kHz.txt";
   -- Input file. Format: x1
   constant sim_in : string (1 to 50) := 
         "simdata\stim_1kHz_17s.txt#########################";
   -- Output file. Format: t  y1  y2
   constant sim_out : string := 
         "simdata\resp_1kHz_16s.txt#########################";
   -- Filter module name
   constant filter_module : string := "ELP12#####";
   -- ADC clock
   constant adc_clock : time := 2 ns;
   
   -- filter coefficients
   shared variable initdone : boolean := false;
   shared variable iir_coeff : iir_record;
   shared variable iir_param : iir_coeffs;

   -- clock signal
   signal clk : std_logic;

   -- filter input value (ideal biquad)
   signal x1 : real := 0.0;
   -- filter output value
   signal y1 : real := 0.0;
   -- history values
   signal h1 : history_array;
   signal h2 : history_array;
   
   -- filter input value (fixed point biquad)
   signal x2 : adc_type := (others => '0');
   -- filter output value
   signal y2 : filter_output_type := (others => '0');
   
   -- difference
   signal dy : real := 0.0;

begin

   ---------------------------------------------------------------------------
   -- Initialization
   ---------------------------------------------------------------------------
   init: process is
      -- filter coeff file
      file coeff : text;
      -- line
      variable l : line;
      -- status
      variable status : file_open_status;
      -- module
      variable module : string (1 to filter_module'length);
      -- string length
      variable modlen : integer := filter_module'length;
      -- good?
      variable good : boolean;
      -- gain values
      variable dcgain, gain, gainerr : real;
      -- printout
      variable s1, s2, s3, s4 : string (1 to 100);
      variable l1, l2, l3, l4 : integer;

      -- skip leading blank characters
      procedure skip_blank (l : inout line) is
         -- character
         variable ch : character;
      begin
         while l'length /= 0 and
               (l(l'left) = ' ' or l(l'left) = character'val(160) or l(l'left) = HT) loop
            read (l, ch);
         end loop;
      end procedure skip_blank;

      -- skip a number field
      procedure skip_number (l : inout line) is
         -- character
         variable ch : character;
      begin
         while l'length /= 0 and
               (l(l'left) >= '0' and l(l'left) <= '9') loop
            read (l, ch);
         end loop;
      end procedure skip_number;

      -- skip a name field
      procedure skip_name (l : inout line) is
         -- character
         variable ch : character;
      begin
         while l'length /= 0 and
                not (l(l'left) = ' ' or l(l'left) = character'val(160) or l(l'left) = HT) loop
            read (l, ch);
         end loop;
      end procedure skip_name;

   begin
      --
      -- read filter coefficients from file
      file_open (status, coeff, path & filter_coeff);
      assert status = open_ok
         report "File open failed for " & filter_coeff severity failure;
      -- prepare filter module string
      for i in 1 to filter_module'length loop
         if filter_module (i) = '#' then
            modlen := i - 1;
            exit;
         end if;
      end loop;
      report "Filter is " & filter_module(1 to modlen) & " from file " & filter_coeff;
      -- read line by line
      while not endfile (coeff) loop
         readline (coeff, l);
         if (l'length < modlen) then
            read (l, module(1 to l'length));
            module(l'length+1 to module'length) := (others => '#');
         else 
            read (l, module(1 to modlen));
            module(modlen+1 to module'length) := (others => '#');
         end if;
         if module = filter_module then
            skip_blank (l); skip_number (l);
            skip_blank (l); skip_number (l);
            skip_blank (l);
            read (l, iir_coeff.num_biquad, good);
            assert good
               report "Unable to read number of biquad sections" severity error;
            skip_blank (l); skip_number (l);
            skip_blank (l); skip_number (l);
            skip_blank (l); skip_name (l);
            skip_blank (l);
            read (l, iir_coeff.gain, good);
            assert good
               report "Unable to read gain" severity error;
            for i in 1 to iir_coeff.num_biquad loop
               skip_blank (l);
               read (l, iir_coeff.biquads(i).a1, good);
               assert good
                  report "Unable to read coefficient a1" severity error;
               skip_blank (l);
               read (l, iir_coeff.biquads(i).a2, good);
               assert good
                  report "Unable to read coefficient a2" severity error;
               skip_blank (l);
               read (l, iir_coeff.biquads(i).b1, good);
               assert good
                  report "Unable to read coefficient b1" severity error;
               skip_blank (l);
               read (l, iir_coeff.biquads(i).b2, good);
               assert good
                  report "Unable to read coefficient b2" severity error;
               if i /= iir_coeff.num_biquad then
                  readline (coeff, l);
               end if;
            end loop;
         end if;
      end loop;
      -- prepare fixed point version
      iir_param.num_biquad := iir_coeff.num_biquad;
      gain := 1.0;
      gainerr := 1.0;
      for i in iir_coeff.num_biquad downto 1 loop
         iir_param.biquads(i).a1 := convert_filter_coeff (-iir_coeff.biquads(i).a1);
         iir_param.biquads(i).a2 := convert_filter_coeff (-iir_coeff.biquads(i).a2);
         iir_param.biquads(i).b1 := convert_filter_coeff (iir_coeff.biquads(i).b1);
         iir_param.biquads(i).b2 := convert_filter_coeff (iir_coeff.biquads(i).b2);
         dcgain := (1.0+iir_coeff.biquads(i).b1+iir_coeff.biquads(i).b2)/
                   (1.0+iir_coeff.biquads(i).a1+iir_coeff.biquads(i).a2);
         report "DC Gain " & real'image (dcgain) & "  gain err " & real'image (gainerr);          
         iir_param.biquads(i).c0 := convert_filter_gain (gainerr/dcgain);
         gain := gain * 2.0**(to_integer(iir_param.biquads(i).c0));
         gainerr := gainerr / dcgain / 2.0**(to_integer(iir_param.biquads(i).c0));
      end loop;
      gain := iir_coeff.gain / gain *
             2.0**(multiplier_coeff_decimal-multiplier_gain_decimal);
      assert abs(gain) < 2.0 report "Gain out of range " & real'image(gain) severity error;
      iir_param.gain := convert_filter_coeff (gain);
      -- print
      report "IIR COEFFICIENTS (REAL)";
      report "biquad num = " & integer'image (iir_coeff.num_biquad);
      report "gain = " & real'image (iir_coeff.gain);
      for i in 1 to iir_coeff.num_biquad loop
         report "a1 = " & real'image (iir_coeff.biquads(i).a1) & "  " &
                "a2 = " & real'image (iir_coeff.biquads(i).a2) & "  " &
                "b1 = " & real'image (iir_coeff.biquads(i).b1) & "  " &
                "b2 = " & real'image (iir_coeff.biquads(i).b2);
      end loop;
      report "IIR COEFFICIENTS (FIXED POINT)";
      report "biquad num = " & integer'image (iir_param.num_biquad);
      signed_to_string (s1, l1, iir_param.gain, multiplier_gain_decimal);
      report "gain = " & s1(1 to l1);
      for i in 1 to iir_param.num_biquad loop
         signed_to_string (s1, l1, iir_param.biquads(i).a1, multiplier_coeff_decimal);
         signed_to_string (s2, l2, iir_param.biquads(i).a2, multiplier_coeff_decimal);
         signed_to_string (s3, l3, iir_param.biquads(i).b1, multiplier_coeff_decimal);
         signed_to_string (s4, l4, iir_param.biquads(i).b2, multiplier_coeff_decimal);
         report "a1 = " & s1(1 to l1) & "  " &
                "a2 = " & s2(1 to l2) & "  " &
                "b1 = " & s3(1 to l3) & "  " &
                "b2 = " & s4(1 to l4) & "  " &
                "c0 = " & integer'image (to_integer(iir_param.biquads(i).c0)) & " (" &
                real'image (2.0**to_integer(iir_param.biquads(i).c0)) & ")";
      end loop;
      file_close (coeff);
      gain := 1.0e-200;
      assert (gain > 0.0) report "real is not double precision!!!";
      initdone := true;
      wait;
   end process init;


   ---------------------------------------------------------------------------
   -- Clock generator
   ---------------------------------------------------------------------------
   clock_gen: process is
      -- filter values
      variable r : real := 0.0;
   begin
      clk <= '1' after adc_clock / 2, '0' after adc_clock;
      wait for adc_clock;
   end process clock_gen;


   ---------------------------------------------------------------------------
   -- Stimulus: Impulse response
   ---------------------------------------------------------------------------
--   stimulus_impulse: process (clk) is
--      -- filter values
--      variable r : real := 0.0;
--   begin
--      if falling_edge (clk) then
--         r := 1.0;
--         x1 <= r;
--         x2 <= convert_adc_value (r, 2.0**(adc_digits-2));
--      end if;
--   end process stimulus_impulse;

   
   ---------------------------------------------------------------------------
   -- Stimulus: read excitation from file
   ---------------------------------------------------------------------------
   stimulus_file: process is
      -- filter output file
      file input : text;
      -- line
      variable l : line;
      -- status
      variable status : file_open_status;
      -- filter values
      variable r : real := 0.0;
      -- string length
      variable len : integer := sim_in'length;
   begin
      for i in 1 to sim_in'length loop
         if sim_in (i) = '#' then
            len := i - 1;
            exit;
         end if;
      end loop;
      -- read filter coefficients from file
      file_open (status, input, path & sim_in(1 to len), read_mode);
      assert status = open_ok
         report "File open failed for " & sim_in(1 to len) severity failure;
      report "Input file is " & sim_in(1 to len);
      -- read data from disk
      while not endfile (input) loop
         readline (input, l);
         read (l, r);
         wait until falling_edge (clk);
         x1 <= r;
         x2 <= convert_adc_value (r, 2.0**(adc_digits-2));
      end loop;
      file_close (input);
      -- done
      wait;
   end process stimulus_file;


   ---------------------------------------------------------------------------
   -- Ideal filter
   ---------------------------------------------------------------------------
   ideal: process (clk) is
      -- filter values
      variable s : real := 0.0;
      -- history values
      variable hist1 : history_array := (others => 0.0);
      variable hist2 : history_array := (others => 0.0);
   begin
      --
      -- run filter
      if rising_edge (clk) then
         ideal_iir (x1, s, hist1, hist2, iir_coeff);
         y1 <= s;
         h1 <= hist1;
         h2 <= hist2;
      end if;
   end process ideal;


   ---------------------------------------------------------------------------
   -- Fixed point filter
   ---------------------------------------------------------------------------
   fixedpoint: process (clk) is
      -- filter values
      variable s : filter_output_type := (others => '0');
      -- history values
      variable hist : iir_history := (others => 
         (x1 => (others => '0'),
          x2 => (others => '0'),
          y1 => (others => '0'),
          y2 => (others => '0')));
      -- overflow
      variable overflow : boolean;
   begin
      --
      -- run filter
      if rising_edge (clk) then
         fixedpoint_iir (x2, s, hist, iir_param, overflow);
         assert not overflow report "Overflow has occured";
         y2 <= s;
      end if;
   end process fixedpoint;


   ---------------------------------------------------------------------------
   -- Compare results
   ---------------------------------------------------------------------------
   compare: process (y1, y2) is
      variable yy2 : real;
      variable delta : real;
   begin
      yy2 := signed_to_real (y2, 10);
      delta := yy2 - y1;
      dy <= delta;
   end process compare;
  
  
   ---------------------------------------------------------------------------
   -- Write results to disk
   ---------------------------------------------------------------------------
   recording: process is
      -- filter output file
      file output : text;
      -- line
      variable l : line;
      -- status
      variable status : file_open_status;
      -- number of data points to skip
      constant skip : integer := 512*1024;
      -- number of data points to keep
      constant keep : integer := 16*512*1024;
      -- recording enabled?
      constant enabled : boolean := true;
      -- string length
      variable len : integer := sim_out'length;
   begin
      if enabled then
         -- skip
         for i in 1 to skip loop 
            wait until falling_edge (clk);
         end loop;
         --
         -- read filter coefficients from file
         for i in 1 to sim_out'length loop
            if sim_out (i) = '#' then
               len := i - 1;
               exit;
            end if;
         end loop;
         file_open (status, output, path & sim_out(1 to len), write_mode);
         assert status = open_ok
            report "File open failed for " & sim_out(1 to len) severity failure;
         report "Output file is " & sim_out(1 to len);
         -- write data to disk
         for i in 0 to keep-1 loop
            wait until falling_edge (clk);
            if i mod out_decimation = 0 then
               --write (l, real(i) * 2.0**(-19), right, 12, 9);
               write (l, y1, right, 20, 14);
               write (l, signed_to_real (y2, adc_filter_shift+filter_out_shift-1), right, 20, 14);
               writeline (output, l);
            end if;
         end loop;
         file_close (output);
      end if;
      -- done
      wait;
   end process recording;
 
end architecture test;
